'''
Create Triqler input files by converting the evidence.txt output file from
MaxQuant.
'''

from __future__ import print_function

import sys
import os
import collections

import numpy as np

from .. import parsers
from ..triqler import __version__, __copyright__
from . import helpers

def main():
  print('Triqler.convert.maxquant version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  convertMqToTriqler(args.file_list_file, args.in_file, args.out_file, params)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  
  requiredNamed = apars.add_argument_group('required arguments')
  
  apars.add_argument('in_file', default=None, metavar = "IN_FILE",
                     help='''MaxQuant evidence.txt file.
                          ''')
  
  requiredNamed.add_argument('--file_list_file', metavar='L', 
                     help='Simple text file with spectrum file names in first column and condition in second column.',
                     required = True)
  
  apars.add_argument('--out_file', default = "triqler_input.tsv", metavar='OUT', 
                     help='''Path to triqler input file (writing in TSV format).
                          ''')
  
  apars.add_argument('--skip_normalization',
                     help='Skip retention-time based intensity normalization.',
                     action='store_true')
  
  apars.add_argument('--skip_mbr_rows',
                     help='Skips the match-between-runs rows in the output.',
                     action='store_true')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['simpleOutputFormat'] = True
  params['skipNormalization'] = args.skip_normalization
  params['skipMBR'] =  args.skip_mbr_rows
  params['plotScatter'] = False
  
  return args, params
  
def convertMqToTriqler(fileListFile, mqEvidenceFile, triqlerInputFile, params):
  fileInfoList = parsers.parseFileList(fileListFile)
  
  peptideToFeatureMap = parseMqEvidenceFile(mqEvidenceFile, fileInfoList, params)
  
  rTimeArrays, factorArrays = helpers.getNormalizationFactorArrays(peptideToFeatureMap, fileInfoList, params)
  
  helpers.writeTriqlerInputFile(triqlerInputFile, peptideToFeatureMap, rTimeArrays, factorArrays, params)

def parseMqEvidenceFile(mqEvidenceFile, fileInfoList, params):
  fileList, _, _, _ = zip(*fileInfoList)
  reader = parsers.getTsvReader(mqEvidenceFile)
  headers = next(reader) # save the header
  headers = list(map(str.lower, headers))
  
  peptCol = headers.index('modified sequence')
  fileCol = headers.index('raw file')
  chargeCol = headers.index('charge')
  intensityCol = headers.index('intensity')
  proteinCol = headers.index('leading proteins')
  scoreCol = headers.index('score')
  rtCol = headers.index('retention time')
  
  fractionCol = headers.index('fraction') if 'fraction' in headers else -1
  experimentCol = headers.index('experiment') if 'experiment' in headers else -1
  
  print("Parsing MaxQuant evidence.txt file")
  peptideToFeatureMap = collections.defaultdict(list)
  for lineIdx, row in enumerate(reader):
    if lineIdx % 500000 == 0:
      print("  Reading line", lineIdx)
    
    proteins = row[proteinCol].split(";")
    
    linkPEP = 0.0
    key = (row[peptCol], row[chargeCol])
    
    if not row[fileCol] in fileList:
      print("Warning: Could not find %s in the specified file list, skipping row" % row[fileCol])
      continue
    
    fileIdx = fileList.index(row[fileCol])
    run, condition, sample, fraction = fileInfoList[fileIdx]
    if fraction == -1 and fractionCol != -1:
      sample, fraction = row[experimentCol], row[fractionCol]
    
    if key in peptideToFeatureMap:
      featureClusterIdx = peptideToFeatureMap[key][0][0].featureClusterId
    else:
      featureClusterIdx = len(peptideToFeatureMap)
    
    if row[intensityCol] == "" or float(row[scoreCol]) <= 0:
      continue
    
    triqlerRow = parsers.TriqlerInputRow(sample, condition, row[chargeCol], lineIdx, linkPEP, featureClusterIdx, np.log(float(row[scoreCol])), float(row[intensityCol]), row[peptCol], proteins)
    peptideToFeatureMap[key].append((triqlerRow, float(row[rtCol]), fraction))
  
  return peptideToFeatureMap

if __name__ == "__main__":
   main()



# coding=utf-8

"""setup.py: setuptools control for triqler."""

import re
from setuptools import setup, find_packages

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('triqler/triqler.py').read(),
    re.M).group(1)

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install triqler
    #

    name='triqler',  # Required
    version=version,  # Required
    packages = ["triqler", "triqler.convert", "triqler.distribution"],
    entry_points = {
        "console_scripts": ['triqler = triqler.triqler:main']
    },

    description='A combined identification and quantification error model of label-free protein quantification',  # Required
    long_description=long_description,  # Optional
    url='https://github.com/statisticalbiotechnology/triqler',  # Optional
    author='Matthew The, Lukas Käll, KTH',  # Optional
    author_email='matthew.the@scilifelab.se',  # Optional
    maintainer='Matthew The, KTH',  # Optional
    maintainer_email='matthew.the@scilifelab.se',  # Optional
    license='Apache',
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    # keywords='sample setuptools development',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    #packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy>=1.12',
        'scipy>=0.17'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'distribution': ['matplotlib'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    #package_data={  # Optional
    #    'sample': ['package_data.dat'],
    #},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],  # Optional

)



from __future__ import print_function

import sys
import csv
import bisect
from collections import defaultdict

import numpy as np

from .. import parsers

def normalizeIntensitiesRtimeBased(clusterQuantExtraFile, clusterQuantExtraNormalizedFile, minRunsObservedIn, plotScatter = False, plotRunningAverage = False):
  featureGroups = parsers.parseFeatureClustersFile(clusterQuantExtraFile)
  factorPairs = getIntensityFactorPairs(featureGroups, sortKey = lambda x : (1, x.fileName, -1.0 * x.intensity, x.rTime), minRunsObservedIn = minRunsObservedIn)
  
  if plotScatter:
    plotFactorScatter(factorPairs)
  
  rTimeFactorArrays = getFactorArrays(factorPairs)
  
  if plotRunningAverage:
    plotFactorRunningAverage(rTimeFactorArrays)
  
  normalizeIntensitiesWithFactorArrays(clusterQuantExtraFile, rTimeFactorArrays, clusterQuantExtraNormalizedFile)

def getIntensityFactorPairs(featureGroups, sortKey, minRunsObservedIn, fraction = 1):
  factorPairs = defaultdict(list)
  for i, featureGroup in enumerate(featureGroups):
    localFactorPairs = dict()
    sortedFeatures = sorted(featureGroup, key = sortKey)
    for row in sortedFeatures:
      rowFraction, fileName, intensity, rTime = sortKey(row)
      intensity *= -1.0
      if fraction == rowFraction and not np.isnan(intensity) and fileName not in localFactorPairs:
        localFactorPairs[fileName] = (np.log2(intensity), rTime)
    
    if len(localFactorPairs) >= minRunsObservedIn:
      keys = sorted(localFactorPairs.keys())
      masterKey = keys[0]
      factor0 = sum([localFactorPairs[masterKey][0] - x[0] for x in localFactorPairs.values()]) / len(keys)
      factorPairs[masterKey].append((localFactorPairs[masterKey][1], factor0))
      for key in keys[1:]:
        factori = factorPairs[masterKey][-1][1] - (localFactorPairs[masterKey][0] - localFactorPairs[key][0])
        factorPairs[key].append((localFactorPairs[key][1], factori))
    #if (i+1) % 100000 == 0:
    #  print("Processing cluster", i+1)
  return factorPairs

# returns running averages of factors
def getFactorArrays(factorPairs, N = 2000):
  factorArrays = defaultdict(list)
  for i, key in enumerate(sorted(factorPairs.keys())):
    print("Calculating normalization factors for run", i+1, "using", len(factorPairs[key]), "precursors")
    #medianFactor = np.median([x[1] for x in factorPairs[key]])
    #print("Calculating normalization factors for run", i+1, "using", len(factorPairs[key]), "precursors", medianFactor)
    factorPairs[key] = sorted(factorPairs[key], key = lambda x : x[0])
    rTimes = [x[0] for x in factorPairs[key]]
    factors = [x[1] for x in factorPairs[key]]
    #factors = [medianFactor for x in factorPairs[key]]
    runningMeans = runningMean(factors, N)
    factorArrays[key] = zip(rTimes, runningMeans)
  return factorArrays

def runningMean(x, N):
  if len(x) <= N:
    return np.array([np.mean(x)]*len(x))
  else:
    cumsum = np.cumsum(np.insert(x, 0, 0))
    rm = (cumsum[N:] - cumsum[:-N]) / N 
    return np.concatenate(([rm[0]]*int(N/2), rm, [rm[-1]]*int(N/2)))
  
def normalizeIntensitiesWithFactorArrays(clusterQuantExtraFile, rTimeFactorArrays, clusterQuantExtraNormalizedFile):
  rTimeArrays, factorArrays = dict(), dict()
  for key in rTimeFactorArrays:
    rTimeArrays[key], factorArrays[key] = zip(*rTimeFactorArrays[key])
  print("Writing", clusterQuantExtraNormalizedFile)
  writer = parsers.getTsvWriter(clusterQuantExtraNormalizedFile)
  precClusters = parsers.parseFeatureClustersFile(clusterQuantExtraFile)
  for i, precCluster in enumerate(precClusters):
    for row in precCluster:
      outRow = list(row[1:])
      outRow[4] = getNormalizedIntensity(rTimeArrays[row.fileName], factorArrays[row.fileName], row.rTime, row.intensity)
      writer.writerow(outRow)
    writer.writerow([])
    if (i+1) % 50000 == 0:
      print("Writing cluster", i+1)

def getNormalizedIntensity(rTimeArray, factorArray, rTime, intensity):
  rTimeIndex = min([bisect.bisect_left(rTimeArray, rTime), len(rTimeArray) - 1])
  return intensity / (2 ** factorArray[rTimeIndex])

def plotFactorScatter(factorPairs):
  from . import scatter
  import matplotlib.pyplot as plt
  scatter.prepareSubplots()
  for i, key in enumerate(sorted(factorPairs.keys())):
    plt.subplot((len(factorPairs)-1) / 4 + 1,4,i+1)
    print("Calculating density", i+1)
    rTimes = [x[0] for x in factorPairs[key] if abs(x[1]) < 2]
    factors = [x[1] for x in factorPairs[key] if abs(x[1]) < 2]
    scatter.plotDensityScatterHist(rTimes, factors, bins = [20,20])
    plt.plot([min(rTimes),max(rTimes)],[0,0],'k:')
    plt.title(key.replace("JD_06232014_","").replace("_","\_"))
    plt.xlabel("Retention time")
    plt.ylabel("log2(int0/int1)")
    plt.ylim([-2,2])
  plt.show()

def plotFactorRunningAverage(factorArrays):
  from . import scatter
  import matplotlib.pyplot as plt
  scatter.prepareSubplots()
  for i, key in enumerate(sorted(factorArrays.keys())):
    plt.subplot((len(factorArrays)-1) / 4 + 1,4,i+1)
    plt.title(key.replace("JD_06232014_","").replace("_","\_"))
    plt.xlabel("Retention time")
    plt.ylabel("log2(int0/int1)")
    plt.ylim([-2,2])
    plt.plot(*zip(*factorArrays[key]))
  plt.show()

if __name__ == "__main__":
   main(sys.argv[1:])
